/*
 *  w3reg by Davide Libenzi ( web pages promoter )
 *  Copyright (C) 2000, 2001  Davide Libenzi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */



#ifdef __LINUX__



#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <sched.h>
#include <pthread.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <dirent.h>
#include <stdarg.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <syslog.h>


#define closesocket                 close
#define stricmp                     strcasecmp
#define strnicmp                    strncasecmp
#define SAIN_Addr(s)                (s).sin_addr.s_addr



#else           // #ifdef __LINUX__
#ifdef WIN32



#include <winsock2.h>
#include <mswsock.h>
#include <windows.h>
#include <windowsx.h>
#include <tchar.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <direct.h>
#include <string.h>
#include <stdlib.h>
#include <process.h>
#include <malloc.h>
#include <ctype.h>
#include <time.h>
#include <stdarg.h>
#include <limits.h>



#define SAIN_Addr(s)                    (s).sin_addr.S_un.S_addr



#else           // #ifdef WIN32

#error System type not defined

#endif          // #ifdef WIN32
#endif          // #ifdef __LINUX__





#define VERSION_STRING              "0.21"






int             iMaxThreads = 32,
                iNumThreads = 0,
                iLogRegs = 0;
char            szEmailAddr[256] = "w3reg@w3reg.net",
                szCountry[128] = "US",
                szIPAddr[128] = "";








char           *StrLTrim(char *pszString)
{

    int             ii;

    for (ii = 0; (pszString[ii] == ' ') || (pszString[ii] == '\t'); ii++);

    if ((ii > 0) && (pszString[ii] != '\0'))
    {
        int             jj;

        for (jj = ii; pszString[jj] != '\0'; jj++)
            pszString[jj - ii] = pszString[jj];

        pszString[jj - ii] = pszString[jj];
    }

    return (pszString);

}




char           *StrRTrim(char *pszString)
{

    int             ii = strlen(pszString) - 1;

    for (; (ii >= 0) && ((pszString[ii] == ' ') || (pszString[ii] == '\t')); ii--)
        pszString[ii] = '\0';


    return (pszString);

}




char           *StrTrim(char *pszString)
{

    return (StrRTrim(StrLTrim(pszString)));

}




char          **AllocReqPkt(char const * pszSite, char const * pszPromote, char const * pszURL)
{

    char          **ppszArgs = (char **) malloc(4 * sizeof(char *));

    if (ppszArgs != NULL)
    {
        ppszArgs[0] = strdup(pszSite);
        ppszArgs[1] = strdup(pszPromote);
        ppszArgs[2] = strdup(pszURL);
        ppszArgs[3] = NULL;
    }

    return (ppszArgs);

}



void            FreeReqPkt(char **ppszArgs)
{

    free(ppszArgs[2]);
    free(ppszArgs[1]);
    free(ppszArgs[0]);
    free(ppszArgs);

}




char           *UniGetString(FILE * pFile, char *pszBuffer, int iBufferSize)
{

    int             ii;

    for (ii = 0; ii < (iBufferSize - 1);)
    {
        int             iChar = fgetc(pFile);

        if (iChar == EOF)
            return (NULL);

        if (iChar == '\r')
        {
            iChar = fgetc(pFile);

            if ((iChar != EOF) && (iChar != '\n'))
                ungetc(iChar, pFile);

            break;
        }
        else if (iChar == '\n')
        {
            iChar = fgetc(pFile);

            if ((iChar != EOF) && (iChar != '\r'))
                ungetc(iChar, pFile);

            break;
        }
        else
            pszBuffer[ii++] = (char) iChar;
    }

    pszBuffer[ii] = '\0';

    return (pszBuffer);

}




int             SplitRequest(char const * pszRequest, char *pszHostName, int *piPortNo,
                        char *pszDocument)
{

    char           *pszEndHost,
                   *pszPort;

    if (strnicmp(pszRequest, "http://", 7) == 0)
        pszRequest += 7;
    else if (strnicmp(pszRequest, "https://", 8) == 0)
        pszRequest += 8;

    if ((pszEndHost = strchr(pszRequest, '/')) == NULL)
        return (-1);

    strncpy(pszHostName, pszRequest, (int) (pszEndHost - pszRequest));
    pszHostName[(int) (pszEndHost - pszRequest)] = '\0';

    if ((pszPort = strrchr(pszHostName, ':')) != NULL)
    {
        *pszPort = '\0';

        *piPortNo = atoi(pszPort + 1);
    }

    strcpy(pszDocument, pszEndHost);

    return (0);

}



char           *QuoteURL(char const * pszURL, char *pszQURL)
{

    int             ii;

    for (ii = 0; *pszURL != '\0'; pszURL++)
    {
        switch (*pszURL)
        {
            case ' ':
            case '\t':
            case ':':
            case '/':
                sprintf(pszQURL + ii, "%%%02X", (int) *pszURL);
                ii += 3;
                break;

            default:
                pszQURL[ii++] = *pszURL;
        }
    }

    pszQURL[ii] = '\0';

    return (pszQURL);

}




int             InstantiateRequest(char *pszDocument, char const * pszIP,
                        char const * pszURL)
{

    char           *pszTag;
    char            szBuffer[1024];

    if ((pszTag = strstr(pszDocument, "<URL>")) != NULL)
    {
        strcpy(szBuffer, pszTag + 5);
        strcpy(pszTag, pszURL);
        strcat(pszTag, szBuffer);
    }

    if ((pszTag = strstr(pszDocument, "<IPADDR>")) != NULL)
    {
        strcpy(szBuffer, pszTag + 9);
        strcpy(pszTag, pszIP);
        strcat(pszTag, szBuffer);
    }

    if ((pszTag = strstr(pszDocument, "<EMAIL>")) != NULL)
    {
        strcpy(szBuffer, pszTag + 8);
        strcpy(pszTag, szEmailAddr);
        strcat(pszTag, szBuffer);
    }

    if ((pszTag = strstr(pszDocument, "<COUNTRY>")) != NULL)
    {
        strcpy(szBuffer, pszTag + 10);
        strcpy(pszTag, szCountry);
        strcat(pszTag, szBuffer);
    }

    return (0);

}



int             SockSend(int SockFD, char *pszBuffer, int iSize)
{

    while (iSize > 0)
    {
        int             iSendSz = send(SockFD, pszBuffer, iSize, 0);

        if (iSendSz < 0)
            return (-1);

        iSize -= iSendSz;
        pszBuffer += iSendSz;
    }

    return (0);

}




int             PromoteURL(char const * pszSite, char const * pszPromote,
                        char const * pszURL)
{

    int             SockFD,
                    iPortNo = 80,
                    iRecvSz,
                    iDumpHttp = iLogRegs;
    unsigned long   ulAddress;
    struct hostent *pHostEnt;
    FILE           *pDumpFile = NULL;
    char           *pszToken;
    struct sockaddr_in Addr;
    char            szHostName[256],
                    szDocument[1024],
                    szDumpFile[512],
                    szQURL[512],
                    szBuffer[2048];

    if (SplitRequest(pszPromote, szHostName, &iPortNo, szDocument) < 0)
        return (-1);

    if (((ulAddress = inet_addr(szHostName)) == INADDR_NONE) &&
            ((pHostEnt = gethostbyname(szHostName)) != NULL))
        ulAddress = *(unsigned long *) pHostEnt->h_addr_list[0];

    if (ulAddress == INADDR_NONE)
        return (-1);

    if ((SockFD = socket(AF_INET, SOCK_STREAM, 0)) == -1)
        return (-1);

    memset(&Addr, 0, sizeof(Addr));
    Addr.sin_family = AF_INET;
    SAIN_Addr(Addr) = ulAddress;
    Addr.sin_port = htons((short int) iPortNo);

    if (connect(SockFD, (struct sockaddr *) & Addr, sizeof(Addr)) < 0)
    {
        closesocket(SockFD);
        return (-1);
    }

    QuoteURL(pszURL, szQURL);

    if (InstantiateRequest(szDocument, (strlen(szIPAddr) > 0) ? szIPAddr : inet_ntoa(Addr.sin_addr),
                    szQURL) < 0)
    {
        closesocket(SockFD);
        return (-1);
    }

    sprintf(szBuffer,
            "GET %s HTTP/1.1\r\n"
            "Host: %s:%d\r\n"
            "Connection: close\r\n"
            "\r\n",
            szDocument, szHostName, iPortNo);

    if (SockSend(SockFD, szBuffer, strlen(szBuffer)) < 0)
    {
        closesocket(SockFD);
        return (-1);
    }

    sprintf(szDumpFile, "%s-%s.tmp", pszSite, szQURL);

    if ((pDumpFile = fopen(szDumpFile, "w+b")) == NULL)
    {
        perror(szDumpFile);
        closesocket(SockFD);
        return (-1);
    }

    while ((iRecvSz = recv(SockFD, szBuffer, sizeof(szBuffer), 0)) > 0)
    {
        szBuffer[iRecvSz] = '\0';

        fputs(szBuffer, pDumpFile);
    }

    closesocket(SockFD);

    fseek(pDumpFile, 0, SEEK_SET);

    if ((UniGetString(pDumpFile, szBuffer, sizeof(szBuffer)) == NULL) ||
            (strtok(szBuffer, " ") == NULL) || ((pszToken = strtok(NULL, " ")) == NULL))
    {
        fclose(pDumpFile);
        remove(szDumpFile);
        return (-1);
    }

    if (atoi(pszToken) != 200)
    {
        fprintf(stderr, "registration error for site \"%s\"  http-error = %d\n",
                pszSite, atoi(pszToken));

        ++iDumpHttp;
    }

    if (iDumpHttp)
    {
        FILE           *pHtmlFile;
        char            szHtmlFile[512];

        while (UniGetString(pDumpFile, szBuffer, sizeof(szBuffer)) != NULL)
            if (strlen(szBuffer) == 0)
                break;

        sprintf(szHtmlFile, "%s-%s.html", pszSite, szQURL);

        if ((pHtmlFile = fopen(szHtmlFile, "wb")) == NULL)
        {
            perror(szHtmlFile);
            fclose(pDumpFile);
            remove(szDumpFile);
            return (-1);
        }

        while (UniGetString(pDumpFile, szBuffer, sizeof(szBuffer)) != NULL)
            fprintf(pHtmlFile, "%s\r\n", szBuffer);

        fclose(pHtmlFile);
    }

    fclose(pDumpFile);
    remove(szDumpFile);

    return (0);

}




int             PromoteProc(char **ppszArgs)
{

    int             iPromoteResult = PromoteURL(ppszArgs[0], ppszArgs[1], ppszArgs[2]);

    --iNumThreads;

    FreeReqPkt(ppszArgs);

    return (iPromoteResult);

}




#ifdef __LINUX__



#define W3REG_CONFIG_FILE               "/etc/w3reg.conf"





int             InitLibrary(void)
{

    return (0);

}



void            CleanupLibrary(void)
{


}



int             PromoteThread(char const * pszSite, char const * pszPromote,
                        char const * pszURL)
{

    pthread_t       TID;
    char          **ppszArgs = AllocReqPkt(pszSite, pszPromote, pszURL);

    ++iNumThreads;

    if (pthread_create(&TID, NULL, (void *(*) (void *)) PromoteProc, (void *) ppszArgs) != 0)
    {
        --iNumThreads;
        FreeReqPkt(ppszArgs);
        return (-1);
    }

    pthread_detach(TID);

    return (0);

}




void            SysSleep(int iTimeout)
{

    sleep(iTimeout);

}




int             SysExistFile(const char *pszFilePath)
{

    return ((access(pszFilePath, F_OK) == 0) ? 1 : 0);

}




#else           // #ifdef __LINUX__
#ifdef WIN32



#define W3REG_CONFIG_FILE               "w3reg.conf"
#define SOCK_VERSION_REQUESTED          MAKEWORD(2, 0)



int             InitLibrary(void)
{

    WSADATA         WSD;

    memset(&WSD, 0, sizeof(WSD));

    if (WSAStartup(SOCK_VERSION_REQUESTED, &WSD))
        return (-1);

    return (0);

}



void            CleanupLibrary(void)
{

    WSACleanup();

}



int             PromoteThread(char const * pszSite, char const * pszPromote,
                        char const * pszURL)
{

    char          **ppszArgs = AllocReqPkt(pszSite, pszPromote, pszURL);
    unsigned int    uThreadId = 0;
    unsigned long   ulThread;

    ++iNumThreads;

    if ((ulThread = _beginthreadex(NULL, 0, (unsigned (__stdcall *) (void *)) PromoteProc,
                            ppszArgs, 0, &uThreadId)) == 0)
    {
        --iNumThreads;
        FreeReqPkt(ppszArgs);
        return (-1);
    }

    CloseHandle((HANDLE) ulThread);

    return (0);

}




void            SysSleep(int iTimeout)
{

    Sleep(iTimeout * 1000);

}




int             SysExistFile(const char *pszFilePath)
{

    return ((_access(pszFilePath, 00) == 0) ? 1 : 0);

}




#else           // #ifdef WIN32

#error System type not defined

#endif          // #ifdef WIN32
#endif          // #ifdef __LINUX__







int             WaitFreeThread(int iMaxThreads)
{

    while (iNumThreads > iMaxThreads)
        SysSleep(1);

    return (iNumThreads);

}





int             MultiPromoteURL(char const * pszSitesFile, char const * pszURL)
{

    int             iCurrLine = 0;
    FILE           *pFile;
    char            szSite[256],
                    szPromote[1024];

    if ((pFile = fopen(pszSitesFile, "rt")) == NULL)
    {
        perror(pszSitesFile);
        return (-1);
    }

    while (fgets(szSite, sizeof(szSite) - 1, pFile) != NULL)
    {
        ++iCurrLine;
        szSite[strlen(szSite) - 1] = '\0';

        StrTrim(szSite);

        if (strlen(szSite) > 0)
        {
            if (fgets(szPromote, sizeof(szPromote) - 1, pFile) == NULL)
                break;

            ++iCurrLine;
            szPromote[strlen(szPromote) - 1] = '\0';

            StrTrim(szPromote);

            if (strlen(szPromote) == 0)
            {
                fprintf(stderr, "syntax error in \"%s\"  line = %d\n",
                        pszSitesFile, iCurrLine);
                fclose(pFile);
                return (-1);
            }

            if (WaitFreeThread(iMaxThreads - 1) >= 0)
                PromoteThread(szSite, szPromote, pszURL);
        }
    }

    fclose(pFile);

    return (0);

}




void            ShowUsage(char const * pszProgName)
{

    fprintf(stderr,
            "W3REG  Version %s by Davide Libenzi\n"
            "       email          = davide_libenzi@mycio.com\n"
            "       home page      = http://www.mycio.com/davidel\n\n"
            "use :  %s  [options]  [urls-file]\n\n"
            "options :\n"
            "       --help         = show this help\n"
            "       --config s     = configuration file\n"
            "       --ipaddr s     = ip address\n"
            "       --threads n    = number of threads\n"
            "       --email s      = user email address\n"
            "       --country s    = user country\n"
            "       --logs         = enable logs\n\n", VERSION_STRING, pszProgName);

}




int             main(int iArgCount, char const * pszArgs[])
{

    int             ii,
                    iExitCode = 0;
    FILE           *pFile;
    char            szConfFile[256] = W3REG_CONFIG_FILE,
                    szURL[1024];

    for (ii = 1; (ii < iArgCount) && (strncmp(pszArgs[ii], "--", 2) == 0); ii++)
    {
        if (strcmp(pszArgs[ii], "--email") == 0)
        {
            if (++ii < iArgCount)
                strcpy(szEmailAddr, pszArgs[ii]);
        }
        else if (strcmp(pszArgs[ii], "--country") == 0)
        {
            if (++ii < iArgCount)
                strcpy(szCountry, pszArgs[ii]);
        }
        else if (strcmp(pszArgs[ii], "--config") == 0)
        {
            if (++ii < iArgCount)
                strcpy(szConfFile, pszArgs[ii]);
        }
        else if (strcmp(pszArgs[ii], "--ipaddr") == 0)
        {
            if (++ii < iArgCount)
                strcpy(szIPAddr, pszArgs[ii]);
        }
        else if (strcmp(pszArgs[ii], "--threads") == 0)
        {
            if (++ii < iArgCount)
                iMaxThreads = atoi(pszArgs[ii]);
        }
        else if (strcmp(pszArgs[ii], "--logs") == 0)
        {
            ++iLogRegs;
        }
        else if (strcmp(pszArgs[ii], "--help") == 0)
        {
            ShowUsage(pszArgs[0]);
            return (0);
        }
        else
        {
            ShowUsage(pszArgs[0]);
            return (1);
        }
    }

    if (!SysExistFile(szConfFile))
    {
        fprintf(stderr, "configuration file not found  \"%s\"\n", szConfFile);
        return (2);
    }

    if (InitLibrary() < 0)
        return (3);

    if (ii < iArgCount)
    {
        if ((pFile = fopen(pszArgs[ii], "rt")) == NULL)
        {
            perror(pszArgs[ii]);
            CleanupLibrary();
            return (4);
        }
    }
    else
        pFile = stdin;


    while (fgets(szURL, sizeof(szURL) - 1, pFile) != NULL)
    {
        szURL[strlen(szURL) - 1] = '\0';

        StrTrim(szURL);

        if ((strlen(szURL) > 0) && (MultiPromoteURL(szConfFile, szURL) < 0))
        {
            iExitCode = 5;
            break;
        }
    }

    if (pFile != stdin)
        fclose(pFile);

    WaitFreeThread(0);

    CleanupLibrary();

    return (iExitCode);

}
